//----------- Copyright  2005-2007, 7thSquad, all rights reserved ------------
//
// This contents is the property of 7thSquad team.
// The contents may be used and/or copied only with the written permission of
// 7thSquad team, or in accordance with the terms and conditions stipulated in
// the agreement/contract under which the contents have been supplied.
//
// Author:	KEXIK
// Purpose:	Natn DLL modul, zskvn ukazatel na tdy (klas. i singl.).
//			Pouze pro Windows.
// Notes: Je ale dleit sledovat co je singleton a neukonovat ho...
//		  Nen to oeteno, muste na to dvat pozor, to snad nen tak tk.
//-----------------------------------------------------------------------------
#pragma once

#include "windows.h"
#include "stdio.h"
#include "symbols.h"
#include <string>
using namespace std;

#define ENTRYPOINT_NAME DllSynapse // nazev entrypointu kazde dll - odviji se od nazvu tymu a projektu
#define ENTRYPOINT_NAME_STR "DllSynapse" // totez jako vyse jen jako retezec

#define MAX_MODULES 30 // maximalni pocet modulu

class ClassHolder;
class CClassManager;

typedef void* (*ClassFactoryFn)(const char* clsName);
typedef void* (*CreateClsFn)();
typedef ClassHolder* (*ModuleEntrypoint)(CClassManager* pClsMgr);

void DbgPrint(WORD objSize, const char* text, ...);

//-----------------------------------------------------------------------------
// CLASS MANAGER
//----------------------------------------------------------------------------- 

class CClassManager
{
public:
	CClassManager();
	void Initialize(const char* szDllPath); // VOLA EXE: najde vsechny knihovny aplikace v zadanem adresari
	void SetClassFactory(ClassFactoryFn fn); // VOLA EXE: nastavi class factory
	bool LoadNewModule(const char* szMouleName); // VOLA EDITOR PRI NACITANI .EXE MODULU HRY
	void Shutdown(); // VOLA EXE: uvolni vsechny singletony a dll moduly
	void* GetClass(const char* className); // vrati instanci tridy podle jejiho identifikatoru
	void PrintFile(const char* data, size_t dataSize);
private:
	ClassHolder* m_pLastHolder; // holdery
	UINT m_nNumSingletons; // pocet jiz nactenych singletonu

	HMODULE m_modules[MAX_MODULES]; // dll moduly
	UINT m_nNumModules; // pocet dll modulu

	char m_szDllPath[MAX_PATH];

	ClassFactoryFn m_clsFactory;
	string m_debugLines;
};

//-----------------------------------------------------------------------------
// POMOCN FUNKCE A DEFINICE
//----------------------------------------------------------------------------- 

void* I_GetClass(const char* szClassName);
void I_Initialize(const char* szDllPath); // POZOR! VOLA POUZE JEDNOU EXE
void I_RegisterModule(const char* szModuleName);
void I_RegisterClassFactory(ClassFactoryFn clsFn);
bool I_LoadNewModule(const char* szMouleName); // POZOR! VOLA EDITOR PRI NACITANI .EXE MODULU HRY
void I_Shutdown(); // POZOR! VOLA POUZE JEDNOU EXE

#define  DLL_EXPORT   extern "C" __declspec( dllexport )

//-----------------------------------------------------------------------------
// DRZAK JEDNE TRIDY (MUZE VYTVARET I VICE INSTANCI DLE TYPU)
//-----------------------------------------------------------------------------
class ClassHolder
{
public:
	ClassHolder(CreateClsFn clsFn, const char* clsName);

	const char* m_clsName;
	CreateClsFn m_clsFn;
	ClassHolder* m_pNext;
	unsigned int m_nID; // poradi inicializace singletonu (pokud je 0 neni singleton nebo byl ukoncen) - ukoncovat se bude od nejvetsiho k nejmensimu
	void* m_pClass; // pokud je trida singleton, bude zde na ni ulozen ukazatel (pokud bude NULL, jeste nebyla vytvorena instance)
	static ClassHolder* g_pLastHolder;
	DWORD m_numInstances; // actual number of class instances
};

//-----------------------------------------------------------------------------
// ZKLADN INTERFACE - VECHNY JSOU OD NJ ODVOZENY
//-----------------------------------------------------------------------------
class IP3DBaseInterface
{
public:
	IP3DBaseInterface(){z__Holder=NULL;};
	virtual ~IP3DBaseInterface() {
		if (z__Holder) 
		{
			DbgPrint(-(WORD)sizeof(this), "DELETE : Deleting class '%s'", z__Holder->m_clsName);
			z__Holder->m_numInstances--;
		}
	};
	virtual bool Init() { return true; };
	void z__SetHolder(ClassHolder* pHolder){ z__Holder=pHolder; };
public:
	ClassHolder* z__Holder;
};

extern CClassManager *g_pClsMgr;

//-----------------------------------------------------------------------------
// MAKRA PRO JEDNODU PRCI S IFACE SYSTMEM (REGISTRACE)
//-----------------------------------------------------------------------------
// Pi vytven nov tdy v projektu, muste vytvoit jej interface (abstraktn tdu) a v cpp souboru
// tdy zaregistrovat danou tdu. To provedete nap. takto:
// REGISTER_CLASS(CP3DSky, "P3DSky"); zaregistruje dve deklarovanou tdu CP3DSky. Potom bude mono zskat
// neomezen mnostv instanc tto tdy dle nzvu "CP3DSky". Pokud chcete umoni vytvet jen
// jednu instanci (singleton) njak tdy pro cel projekt, pouijte REGISTER_SINGLE_CLASS.
// Duleit informace:
// - Singletony se samy vytvo pouze jednou, jsou-li poteba. Pi ukonen je poteba zavolat I_UnloadSingletons().
// - Klasick tdy se vytvo a zsk se ukazatel. Pi ukonovn je teba takovto tdy ukonit pomoc delete.

// Zaregistruje tdu a nastav vytvec funkci na zadanou "createClsFn".
// Pozor! Funkce se mus sama postarat o vytvoen instance tdy a zavoln INIT()!!!
#define REGISTER_CLASS_FN(className, callNameStr, createClsFn) \
	static ClassHolder __g_cls##className##_holder(createClsFn, callNameStr);

// Zaregistruje tdu a vytvo pro ni vytvec funkci.
#define REGISTER_CLASS(className, callNameStr) \
	extern ClassHolder __g_cls##className##_holder;\
	static void* __Create##className##Class() {className* pCls=new className;\
	if (pCls->Init()){ __g_cls##className##_holder.m_numInstances++; pCls->z__SetHolder(&__g_cls##className##_holder); return (void*)pCls; } else {delete pCls; return NULL;} }\
	REGISTER_CLASS_FN(className, callNameStr, __Create##className##Class);

// Zaregistruje jedinenou tdu (singleton) z ukazatele "globalClassVar". Tda se mus o sebe
// postarat sama (vytvoit se a ukonit se). globalClassVar by proto mla bt promnn typu
// dan tdy. Potom se o ukonen postar systm sm.
// Take se vytvo teba CMojeTrida g_singl; REGISTER_SINGLE_CLASS_G(-, -, g_singl);
#define REGISTER_SINGLE_CLASS_G(className, callNameStr, globalClassVar) \
	static BYTE __g##className##_bInitState=0; /*stav voln Init(): 0=Create jet nebylo volno, 1=Init ok, 2=err*/\
	static void* __Create##className##Class() {if(!__g##className##_bInitState)\
	{__g##className##_bInitState = globalClassVar.Init() ? 1 : 2;}\
	return __g##className##_bInitState==1 ? (void*)&globalClassVar : NULL;}\
	static ClassHolder __g_cls##className##_holder(__Create##className##Class, callNameStr);

// Zaregistruje jedinenou tdu (singleton) a vytvo globn ukazatel na tuto tdu.
// Takto vytvoen singleton se postar o vytvoen a ukonen instance sm.
#define REGISTER_SINGLE_CLASS(className, callNameStr) \
	static className* __g_##className##_singleton=NULL;\
	extern ClassHolder __g_cls##className##_holderS;\
	static bool __g##className##_bSuccess=false;\
	static void* __Create##className##ClassS() {\
		if (!__g_##className##_singleton) {\
			__g_##className##_singleton = new className;\
			__g##className##_bSuccess = __g_##className##_singleton->Init();\
			__g_cls##className##_holderS.m_nID = 1;\
		}\
		if (__g_##className##_singleton && !__g##className##_bSuccess) {delete __g_##className##_singleton;\
		__g_##className##_singleton=NULL;}\
		return __g##className##_bSuccess ? (void*)__g_##className##_singleton : NULL;}\
	static ClassHolder __g_cls##className##_holderS(__Create##className##ClassS, callNameStr);